(*  Title:      HOL/TPTP/Satallax.thy
    Author:     Nik Sultana, Cambridge University Computer Laboratory

Theory support for reconstructing proofs generated by Satallax.

The contents of this file are based on the file "Satallax.thy" which is
distributed with Satallax version "2.7-isar". Both files validate in
Isabelle/HOL the inference rules used by Satallax, but do it in slightly
different ways, owing to the way the validated inference rules are used
subsequently.
*)

theory Satallax
imports Hilbert_Choice
begin

(*NOTE universe Prop is replaced by type bool*)
type_synonym Prop = bool ("Prop")
type_synonym o = bool ("o")
typedecl i(*FIXME is this used?*)

lemma TNegAll :
  "\<And> p R sk::'X. ~(! x::'X.  p x)
   \<Longrightarrow> sk = (SOME x. ~ p x)
   \<Longrightarrow> (\<lbrakk>~(! x::'X. p x); ~p sk\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
  apply (subgoal_tac "~(! x::'X. p x) \<Longrightarrow> ~p sk")
  apply simp
  apply simp
  apply (erule someI_ex)
done

(*Similar to TNegAll, but used when "x" isn't free in "p"*)
lemma TNegAll_notfree :
  "\<And> p R . ~(! x::'X.  p)
   \<Longrightarrow> (\<lbrakk>~(! x::'X. p); ~p\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TNegImp :
  "\<not> (X \<longrightarrow> Y)
   \<Longrightarrow> (\<lbrakk>\<not> (X \<longrightarrow> Y); X; \<not> Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

(*TAnd is similar to conjE, but retains the
  "X & Y" formula.*)
lemma TAnd :
  "X \<and> Y
   \<Longrightarrow> (\<lbrakk>X \<and> Y; X; Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma pre_eq_ind_r :
  "! (x :: 'ty1_) (P :: 'ty1_ => Prop). (P x) --> (! y . y = x --> (P y))"
by blast

lemma chain_hypothesis :
  "P \<Longrightarrow>
   Q \<Longrightarrow>
   (\<lbrakk>Q; P\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemmas eq_ind_r = chain_hypothesis[OF pre_eq_ind_r]

lemma TabAll:
  "\<And>y. ! x. P x \<Longrightarrow> (\<lbrakk>! x. P x; P y\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"
by blast

lemma TImp :
  "X \<longrightarrow> Y \<Longrightarrow> (~X \<Longrightarrow> R) \<Longrightarrow> (Y \<Longrightarrow> R) \<Longrightarrow> R"
by blast

lemma TabConflict :
  "\<lbrakk>X = Y; X \<noteq> Y\<rbrakk> \<Longrightarrow> (P \<Longrightarrow> R) \<Longrightarrow> R"
  "\<lbrakk>~ (X \<noteq> Y); X \<noteq> Y\<rbrakk> \<Longrightarrow> (P \<Longrightarrow> R) \<Longrightarrow> R"
  "\<lbrakk>X; ~X\<rbrakk> \<Longrightarrow> (P \<Longrightarrow> R) \<Longrightarrow> R"
by auto

lemma TabDec :
  "X Y \<noteq> X Z \<Longrightarrow> (\<lbrakk>X Y \<noteq> X Z; Y \<noteq> Z\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"
by blast

lemma TNAnd :
  "~(X & Y) \<Longrightarrow> (~X \<Longrightarrow> R) \<Longrightarrow> (~Y \<Longrightarrow> R) \<Longrightarrow> R"
by blast

lemma TOr :
  "X \<or> Y \<Longrightarrow> (X \<Longrightarrow> R) \<Longrightarrow> (Y \<Longrightarrow> R) \<Longrightarrow> R"
by blast

lemma TNor :
  "~(X \<or> Y) \<Longrightarrow> (\<lbrakk>~(X \<or> Y); ~X; ~Y\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"
by blast

lemma TEx :
  "\<And> p R sk::'X. Ex p
   \<Longrightarrow> sk = (SOME x. p x)
   \<Longrightarrow> (\<lbrakk>Ex p; p sk\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
  apply (subgoal_tac "? x::'X. p x \<Longrightarrow> p sk")
  apply simp
  apply simp
  apply (erule someI_ex)
done

(*Similar to TEx, but used when "x" isn't free in "p"*)
lemma TEx_notfree :
  "\<And> p R. Ex p
   \<Longrightarrow> (\<And> sk. \<lbrakk>Ex p; p sk\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TNegEx :
  "\<And> y. ~ (? x::'X.  p x)
   \<Longrightarrow> (\<lbrakk>~ (? x::'X.  p x); ~ p y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

declare [[show_types]]

lemma TBQ :
  "(X = Y)
   \<Longrightarrow> (\<lbrakk>X = Y; X; Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> (\<lbrakk>X = Y; ~X; ~Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

(*NOTE <> becomes ~=*)
lemma TBE :
  "(X \<noteq> Y)
   \<Longrightarrow> (\<lbrakk>X \<noteq> Y; X; ~Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> (\<lbrakk>X \<noteq> Y; ~X; Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

(*NOTE <--> becomes =*)
lemma TIff :
  "(X \<longleftrightarrow> Y)
   \<Longrightarrow> (\<lbrakk>X \<longleftrightarrow> Y; X; Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> (\<lbrakk>X \<longleftrightarrow> Y; ~X; ~Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TNIff :
  "~(X \<longleftrightarrow> Y)
   \<Longrightarrow> (\<lbrakk>~(X \<longleftrightarrow> Y); X; ~Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> (\<lbrakk>~(X \<longleftrightarrow> Y); ~X; Y\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TFQ :
  "s = t
   \<Longrightarrow> (\<lbrakk>s = t; ! x::'Sig.  s x = (t x :: 'Tau)\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TFE :
  "s \<noteq> t
   \<Longrightarrow> (\<lbrakk>s \<noteq> t; ~(! x::'Sig.  s x = (t x :: 'Tau))\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TCON :
  "\<lbrakk>((s::'Sig) = t); (u ~= v)\<rbrakk> \<Longrightarrow>
   (\<lbrakk>((s::'Sig) = t); (u ~= v)\<rbrakk>
     \<Longrightarrow> \<lbrakk>s ~= u; t ~= u\<rbrakk>
     \<Longrightarrow> R) \<Longrightarrow>
   (\<lbrakk>((s::'Sig) = t); (u ~= v)\<rbrakk>
     \<Longrightarrow> \<lbrakk>s ~= v; t ~= v\<rbrakk>
     \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma Ttrans :
  "\<And> s t u.
   \<lbrakk>s = t; t = u\<rbrakk>
   \<Longrightarrow> (\<lbrakk>s = t; t = u\<rbrakk>
        \<Longrightarrow> \<lbrakk>s = u\<rbrakk>
        \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TDec :
  "(p :: 'Sig => 'Tau) s ~= q t
   \<Longrightarrow> (\<lbrakk>(p :: 'Sig => 'Tau) s ~= q t; ((p ~= q) | (s ~= t))\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TDN :
  "~~P
   \<Longrightarrow> (\<lbrakk>~~P; P\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TMat :
  "\<lbrakk>p s; ~ q t\<rbrakk>
   \<Longrightarrow> (\<lbrakk>p s; ~ q t\<rbrakk>
        \<Longrightarrow> p ~= q
        \<Longrightarrow> R)
   \<Longrightarrow> (\<lbrakk>p s; ~ q t\<rbrakk>
        \<Longrightarrow> s ~= t
        \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

(*like TMat, but where we can skip the step
  of refuting that (p \<noteq> p)*)
lemma TMat3 :
  "\<lbrakk>p s; ~ p t\<rbrakk>
   \<Longrightarrow> (\<lbrakk>p s; ~ p t\<rbrakk>
        \<Longrightarrow> s ~= t
        \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TCut :
  "\<And> P R. P
   \<Longrightarrow> (\<lbrakk>P; Q\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> (\<lbrakk>P; ~Q\<rbrakk> \<Longrightarrow> R)
   \<Longrightarrow> R"
by blast

lemma TabRefl :
  "\<lbrakk>~(X = X)\<rbrakk> \<Longrightarrow> (P \<Longrightarrow> R) \<Longrightarrow> R"
by blast

lemma TabFalse :
  "False \<Longrightarrow> (P \<Longrightarrow> R) \<Longrightarrow> R"
by blast

lemma TabDelta :
  "P \<Longrightarrow> (P \<Longrightarrow> R) \<Longrightarrow> R"
by blast

(*NOTE use Eps instead of Sepsilon*)
lemma pre_TSeps :
  "((p :: 'A => Prop) (Eps p) --> False) --> (! x.  ~ p x)"
by (blast intro: someI)
lemmas TSeps = chain_hypothesis[OF pre_TSeps]

lemma pre_TSeps' :
  "! (eps:: ('A=>Prop)=>'A) (p::'A=>Prop).  (! (q::'A=>Prop) (y::'A).  q y --> q (eps q)) --> (p (eps p) --> False) --> (! x.  ~ p x)"
by (blast intro: someI)
lemmas TSeps' = chain_hypothesis[OF pre_TSeps']

lemma pre_TSeps'' :
  "! (eps:: ('A=>Prop)=>'A) (p::'A=>Prop).   (! (q::'A=>Prop).  ~ (! (y::'A). ~ q y) --> q (eps q)) --> (p (eps p) --> False) --> (! x.  ~ p x)"
by (blast intro: someI)
lemmas TSeps'' = chain_hypothesis[OF pre_TSeps'']

lemma pre_TSeps''' :
  "! (eps:: ('A=>Prop)=>'A) (p::'A=>Prop). (! (q::'A=>Prop).  (? y::'A.  q y) --> q (eps q)) --> (p (eps p) --> False) --> (! x.  ~ p x)"
by (blast intro: someI)
lemmas TSeps''' = chain_hypothesis[OF pre_TSeps''']


(*Lemmas outside Satallax's calculus*)

(*lemma eq_ind_r:
  "! (x :: 'ty1_) (P :: 'ty1_ => Prop). (P x) --> (! y . y = x --> (P y))"
by blast
-- handled this earlier*)

lemma pre_eq_ind:
  "! (x :: 'ty1_) (P :: 'ty1_ => Prop). (P x) --> (! y . x = y --> (P y))"
by blast
lemmas eq_ind = chain_hypothesis[OF pre_eq_ind]


(* Originally from stt.v *)

(* Implicit Arguments prop_ext [A B]. *)
lemma pre_prop_ext :
  "((A :: o) = B) --> (A = B)"
by blast
lemmas prop_ext = chain_hypothesis[OF pre_prop_ext]

(* Implicit Arguments eq_ind_l [A x y]. *)
lemma pre_eq_ind_l:
  "! (P :: 'ty1_ => Prop). (P x) --> (x = y --> (P y))"
by blast
lemmas eq_ind_l = chain_hypothesis[OF pre_eq_ind_l]

(*** Two variants of transitivity of equality. ***)
lemma pre_eq_trans2 :
  "! x y (z :: 'A). x = y --> z = y --> x = z"
by blast
lemmas eq_trans2 = chain_hypothesis[OF pre_eq_trans2]

lemma pre_eq_trans3 :
  "! x y (z :: 'A). y = x --> y = z --> x = z"
by blast
lemmas eq_trans3 = chain_hypothesis[OF pre_eq_trans3]


(*** Logical Identities ***)

lemma pre_eq_neg_neg_id : "(% x :: Prop. ~~x) = (% x :: Prop. x)" by blast
lemma pre_eq_true : "True = (~ False)" by blast
lemma pre_eq_and_nor : "(op &) = (% x (y :: Prop). ~(~x | ~y))" by blast
lemma pre_eq_or_nand : "(op | ) = (% x (y :: Prop). ~(~x & ~y))" by blast
lemma pre_eq_or_imp : "(op | ) = (% x (y :: Prop). (~ x) --> y)" by blast
lemma pre_eq_and_imp : "(op &) = (% x (y :: Prop). ~ (x --> (~ y)))" by blast
lemma pre_eq_imp_or : "(% x (y :: Prop). (x --> y)) = (% x (y :: Prop). ((~x) | y))" by blast

lemmas eq_neg_neg_id = chain_hypothesis[OF pre_eq_neg_neg_id]
lemmas eq_true = chain_hypothesis[OF pre_eq_true]
lemmas eq_and_nor = chain_hypothesis[OF pre_eq_and_nor]
lemmas eq_or_nand = chain_hypothesis[OF pre_eq_or_nand]
lemmas eq_or_imp = chain_hypothesis[OF pre_eq_or_imp]
lemmas eq_and_imp = chain_hypothesis[OF pre_eq_and_imp]
lemmas eq_imp_or = chain_hypothesis[OF pre_eq_imp_or]

syntax
  "=" :: "[bool, bool] => bool"  ("<->" 50)

lemma pre_eq_iff :
  "(% s t. s <-> t) = (% s t. s = t)"
by blast
lemmas eq_iff = chain_hypothesis[OF pre_eq_iff]

lemma pre_eq_sym_eq :
  "(% s (t :: 'sig). s = t) = (% s (t :: 'sig). t = s)"
by blast
lemmas eq_sym_eq = chain_hypothesis[OF pre_eq_sym_eq]

lemma pre_eq_forall_nexists :
  "(% (f :: 'A => Prop). ! x. f x) = (% (f :: 'A => Prop). ~ (? x. ~f x))"
by blast
lemmas eq_forall_nexists = chain_hypothesis[OF pre_eq_forall_nexists]

lemma pre_eq_exists_nforall :
  "(% (f :: 'A => Prop). ? x. f x) = (% (f :: 'A => Prop). ~ (! x. ~ f x))"
by blast
lemmas eq_exists_nforall = chain_hypothesis[OF pre_eq_exists_nforall]

lemma pre_eq_leib1 : "(% s (t :: 'A). ! (p :: 'A => Prop). (p s --> p t)) = (% s (t :: 'A ). s = t)"
  apply (rule ext)
  apply (rule ext)
  apply (rule iffI)
  defer
  apply simp
  apply (erule_tac x = "% z. s = z" in allE)
  apply blast
  done
lemmas eq_leib1 = chain_hypothesis[OF pre_eq_leib1]

lemma pre_eq_leib2 : "(% s (t :: 'A). ! (p :: 'A => Prop). (~ p s) --> ~ p t) = (% s (t :: 'A ). s = t)"
  apply (rule ext)
  apply (rule ext)
  apply (rule iffI)
  defer
  apply simp
  apply (erule_tac x = "% z. t = z" in allE)
  apply blast
done
lemmas eq_leib2 = chain_hypothesis[OF pre_eq_leib2]

lemma pre_eq_leib3 : "(% s (t :: 'A). ! (r :: 'A => 'A => Prop). (! x. r x x) --> r s t) = (% s (t :: 'A). s = t)"
  apply (rule ext)
  apply (rule ext)
  apply (rule iffI)
  defer
  apply simp
  apply (erule_tac x = "op =" in allE)
  apply (erule mp)
  apply (rule allI)
  apply (rule refl)
done
lemmas eq_leib3 = chain_hypothesis[OF pre_eq_leib3]

lemma pre_eq_leib4 : "(% s (t :: 'A). ! (r :: 'A => 'A => Prop). (~ r s t) --> ~ (! x. r x x) ) = (% s (t :: 'A). s = t)"
  apply (rule ext)
  apply (rule ext)
  apply (rule iffI)
  defer
  apply blast
  apply (erule_tac x = "op =" in allE)
  apply blast
done
lemmas eq_leib4 = chain_hypothesis[OF pre_eq_leib4]

(*** eta ***)
lemma pre_eq_eta :
  "(% (f :: 'A => 'B) (x :: 'A). f x) = (% (f :: 'A => 'B). f)"
by blast
lemmas eq_eta = chain_hypothesis[OF pre_eq_eta]


(*from Coq's library*)
lemma pre_eq_sym :
  "! x y. (x :: 'ty1_) = y --> y = x"
by blast
lemmas eq_sym = chain_hypothesis[OF pre_eq_sym]

(*For rewriting.*)
lemma tab_rew :
  "H0 = H1 \<Longrightarrow> H0 \<Longrightarrow>
   (\<lbrakk>H0; H1\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"
by blast

end