(*  Title:      HOL/TPTP/Satallax_Reconstruct_Test.thy
    Author:     Nik Sultana, Cambridge University Computer Laboratory

Testing of Satallax TPTP proof reconstruction.
Based on and extends TPTP_Proof_Reconstruction_Test.

FIXME some functionality in TPTP_Proof_Reconstruction_Test is duplicated.
*)

theory Satallax_Reconstruct_Test
imports
  Satallax_Reconstruct
(*  TPTP_Proof_Reconstruction_Test*)
  TPTP_Test

begin

declare [[show_types]]
declare [[show_sorts]]
declare [[show_consts]]

ML {*
  (Options.default_put_bool @{system_option ML_exception_trace} true;
   default_print_depth 200;  (* FIXME proper ML_print_depth within context!? *)
   PolyML.Compiler.maxInlineSize := 0)
*}

ML {*
reconstruct_satallax
  (Path.explode "$THF_PROOFS/SYO223^5.p")
  @{theory}
*}

ML {*
(*Based on contents of TPTP_Proof_Reconstruction_Test*)
val probs =
  "$THF_PROOFS/SYO223^5.proof"
  |> Path.explode
  |> single

val prob_names =
  probs
  |> map (Path.base #> Path.implode #> TPTP_Problem_Name.Nonstandard)
*}

ML {*
(*FIXME DRY code from TPTP_Proof_Reconstruction_Test*)

  (*this version of "reconstruct" builds theorems, instead of lists of reconstructed inferences*)
  (*default timeout is 1 min*)
  fun reconstruct timeout file thy =
    let
      val timer = Timer.startRealTimer ()
      val ctxt = Proof_Context.init_global thy
    in
      TimeLimit.timeLimit (Time.fromSeconds (if timeout = 0 then 60 else timeout))
       (fn file =>

        (can (fn file =>
           if exists (fn x =>
             Path.base file = x)
           (map (Path.base o Path.explode)
             [(*These problems hang the test*)
              "SYO268^5.p",
              "LCL731^5.p"
           ])
           then error "excluded"
           else reconstruct_satallax file thy) file)
       |> (fn s => report ctxt (Path.print file ^ " === " ^ Bool.toString s ^
             " t=" ^ (Timer.checkRealTimer timer |> Time.toMilliseconds |> PolyML.makestring))))
       file
    end
*}

ML {*
reconstruct_satallax
  (Path.explode "$THF_PROOFS/SEV159^5.p")
  @{theory}
*}

ML {*
  fun reconstruction_test timeout ctxt =
    test_fn ctxt
     (fn file => reconstruct timeout file (Proof_Context.theory_of ctxt))
     "reconstructor"
     ()
*}

ML "tptp_probs_dir"

ML {*
val tptp_probs_dir =
  "$THF_PROOFS"
  |> Path.explode
  |> Path.expand
*}

declare [[
  tptp_trace_reconstruction = false,
  tptp_test_all = true,
  tptp_test_timeout = 10,
  tptp_max_term_size = 0
]]

ML {*
  if test_all @{context} then
    (report @{context} "Reconstructing proofs";
    S timed_test (reconstruction_test (get_timeout @{context})) @{context} (TPTP_Syntax.get_file_list tptp_probs_dir))
  else ()
*}
